/* xscreensaver, Copyright (c) 1997, 1998, 2002 Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 *
 * Concept snarfed from Michael D. Bayne in
 * http://www.go2net.com/internet/deep/1997/04/16/body.html
 */

#include "screenhackff.h"
#include <stdio.h>
#include <time.h>
#include <sys/time.h>

static XColor colors[255];
static int ncolors = 0;
static int max_depth = 0;
static int min_height = 0;
static int min_width = 0;

static void
deco (Display *dpy,
      Window window,
      Colormap cmap,
      GC fgc, GC bgc,
      int x, int y, int w, int h, int depth)
{
  if (((ya_random() % max_depth) < depth) || (w < min_width) || (h < min_height))
    {
      if (!mono_p)
	{
	  static int current_color = 0;
	  if (++current_color >= ncolors)
	    current_color = 0;
	  XSetForeground(dpy, bgc, colors[current_color].pixel);
	}
      XFillRectangle (dpy, window, bgc, x, y, w, h);
      XDrawRectangle (dpy, window, fgc, x, y, w, h);
    }
  else
    {
      if (ya_random() & 1)
	{
	  deco (dpy, window, cmap, fgc, bgc, x, y, w/2, h, depth+1);
	  deco (dpy, window, cmap, fgc, bgc, x+w/2, y, w/2, h, depth+1);
	}
      else
	{
	  deco (dpy, window, cmap, fgc, bgc, x, y, w, h/2, depth+1);
	  deco (dpy, window, cmap, fgc, bgc, x, y+h/2, w, h/2, depth+1);
	}
    }
}


char *progclass = "Deco";
int progtype = FF_SOURCE;

char *defaults [] = {
  ".background:		black",
  ".foreground:		white",
  "*maxDepth:		12",
  "*minWidth:		20",
  "*minHeight:		20",
  "*cycle:		False",
  "*delay:		500",
  "*ncolors:		64",
  0
};

FFParam ff_params [] = {
    {"*delay", 0, 1000, 500, FF_TYPE_STANDARD},
    {"*ncolors", 1, 255, 64, FF_TYPE_STANDARD},
    {"*maxDepth", 1, 40, 12, FF_TYPE_STANDARD},
    {"*minWidth", 1, 100, 20, FF_TYPE_STANDARD},
    {"*minHeight", 1, 100, 20, FF_TYPE_STANDARD},
    {0, 0, 0, 0, 0}
};

XrmOptionDescRec options [] = {
  { "-max-depth",	".maxDepth",	XrmoptionSepArg, 0 },
  { "-min-width",	".minWidth",	XrmoptionSepArg, 0 },
  { "-min-height",	".minHeight",	XrmoptionSepArg, 0 },
  { "-delay",		".delay",	XrmoptionSepArg, 0 },
  { "-ncolors",		".ncolors",	XrmoptionSepArg, 0 },
  { "-cycle",		".cycle",	XrmoptionNoArg, "True" },
  { "-no-cycle",	".cycle",	XrmoptionNoArg, "False" },
  { "-cycle-delay",	".cycleDelay",	XrmoptionSepArg, 0 },
  { 0, 0, 0, 0 }
};

unsigned deco_instantiate(Display *dpy, Window window);
unsigned deco_deinstantiate(void);

unsigned (*screenhack_init)(void) = NULL;
unsigned (*screenhack_deinit)(void) = NULL;
unsigned (*screenhack_instantiate)(Display *, Window) = deco_instantiate;
unsigned (*screenhack_deinstantiate)(void) = deco_deinstantiate;

static unsigned width, height;
static Colormap colormap;
static GC fgc, bgc;
static Display *_dpy;
static Visual *visual;

unsigned deco_instantiate(Display *dpy, Window window)
{
    int screen;
    int dummy;
    void *pdummy = &dummy;
    XGCValues gcv;
    
    _dpy = dpy;
    XGetGeometry(dpy, window, pdummy, pdummy, pdummy, &width, &height, pdummy, pdummy);

    screen = DefaultScreen(dpy);
    visual = DefaultVisual(dpy, screen);
    colormap = XDefaultColormap(dpy, screen);

    gcv.foreground = get_pixel_resource("foreground", "Foreground",
				      dpy, colormap);
    
    fgc = XCreateGC(dpy, window, GCForeground, &gcv);
    
    gcv.foreground = get_pixel_resource("background", "Background",
				      dpy, colormap);
    bgc = XCreateGC(dpy, window, GCForeground, &gcv);

    return FF_SUCCESS;
}

unsigned deco_deinstantiate(void)
{
    XFreeGC(_dpy, bgc);
    XFreeGC(_dpy, fgc);
    return FF_SUCCESS;
}

void
screenhack_frame(Display *dpy, Window window)
{
    static unsigned last_draw = 0;
    int tick;
    int delay;
    Bool writable;
    
    delay = get_integer_resource ("delay", "Integer");
    
    tick = gettickcount();
    if (delay)
    {
	if (last_draw && ((tick - last_draw) < delay))
	    return;
	last_draw = tick;
    }

    max_depth = get_integer_resource ("maxDepth", "Integer");
    if (max_depth < 1) max_depth = 1;
    else if (max_depth > 1000) max_depth = 1000;
    
    min_width = get_integer_resource ("minWidth", "Integer");
    if (min_width < 2) min_width = 2;
    min_height = get_integer_resource ("minHeight", "Integer");
    if (min_height < 2) min_height = 2;
    
    ncolors = get_integer_resource ("ncolors", "Integer");

    make_random_colormap (dpy, visual, colormap, colors, &ncolors,
			False, True, &writable, True);
    
    if (ncolors <= 2)
	mono_p = True;
    else
	mono_p = False;

    if (!mono_p)
    {
	GC tmp = fgc;
	fgc = bgc;
	bgc = tmp;
    }

    XFillRectangle(dpy, window, bgc, 0, 0, width, height);
    deco (dpy, window, colormap, fgc, bgc, 0, 0, width, height, 0);
    XSync (dpy, False);
    
    free_colors (dpy, colormap, colors, ncolors);
}
