/* munch.c
 * A munching squares implementation for X
 * Tim Showalter <tjs@andrew.cmu.edu>
 * 
 * Copyright 1997, Tim Showalter
 * Permission is granted to copy, modify, and use this as long
 * as this notice remains intact.  No warranties are expressed or implied.
 * CMU Sucks.
 * 
 * Some code stolen from / This is meant to work with
 * xscreensaver, Copyright (c) 1992, 1995, 1996
 * Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

/* Munching Squares is this simplistic, silly screen hack (according
   to HAKMEM, discovered by Jackson Wright in 1962) where you take
   Y = X XOR T and graph it over and over.  According to HAKMEM, it 
   takes 5 instructions of PDP-1 assembly.  This is a little more
   complicated than that, mostly X's fault, but it does some other
   random things.

   http://www.inwap.com/pdp10/hbaker/hakmem/hacks.html#item146
 */

#include <math.h>
/*#include <assert.h>*/
#include "screenhackff.h"

/* flags for random things.  Must be < log2(random's maximum), incidentially.
 */
#define SHIFT_KX (0x01)
#define SHIFT_KT (0x02)
#define SHIFT_KY (0x04)
#define GRAV     (0x08)

char *progclass = "Munch";
int progtype = FF_SOURCE;

static unsigned munch_instantiate(Display *, Window);
static unsigned munch_deinstantiate(void);

unsigned (*screenhack_init)(void) = NULL;
unsigned (*screenhack_deinit)(void) = NULL;
unsigned (*screenhack_instantiate)(Display *, Window) = munch_instantiate;
unsigned (*screenhack_deinstantiate)(void) = munch_deinstantiate;

FFParam ff_params [] = {
    {"*delay", 1, 50, 10, FF_TYPE_STANDARD},
    {"*clear", 1, 200, 50, FF_TYPE_STANDARD},
    {"*shift", 0, 1, 1, FF_TYPE_BOOLEAN},
    {"*xor", 0, 1, 1, FF_TYPE_BOOLEAN},
    {0, 0, 0, 0, 0}
};

char *defaults [] = {
    ".background:	black",
    ".foreground:	white",
    "*delay:	        10",
    "*hold:		100000",
    "*clear:		50",
    "*logminwidth:	7",
    "*shift:		True",
    "*xor:		True",
    0
};

XrmOptionDescRec options [] = {
    { "-delay",		".delay",	XrmoptionSepArg, 0 },
    { "-hold",		".hold",	XrmoptionSepArg, 0 },
    { "-clear",         ".clear",       XrmoptionSepArg, "true" },
    { "-shift",         ".shift",       XrmoptionNoArg, "true" },
    { "-no-shift",	".shift",       XrmoptionNoArg, "false" },
    { "-logminwidth",	".logminwidth", XrmoptionSepArg, 0 },
    { "-xor",           ".xor",		XrmoptionNoArg, "true" },
    { "-no-xor",	".xor",		XrmoptionNoArg, "false" },
    { 0, 0, 0, 0 }
};


static GC gc, cleargc;
static Display *_dpy;
static unsigned xgwa_width, xgwa_height;
static Colormap xgwa_colormap;

/* only configurable things right now */
static int delay, hold, clear, logminwidth, shiftk, xor;

static XPoint *points;

static void munchX(Display *dpy, Window w,
    int t, int width, int atX, int atY, int kX, int kT, int kY, int grav)
{
    int x, y;
    
    for(x = 0; x < width; x++) {
	/* figure out the next point */
	y = ((x ^ ((t + kT) % width)) + kY) % width;

	points[x].x = ((x + kX) % width) + atX;
	points[x].y = (grav ? y + atY : atY + width - 1 - y);
    }
    /*XDrawPoint(dpy, w, gc, drawX, drawY);*/
    XDrawPoints(dpy, w, gc, points, width, CoordModeOrigin);
}

static int munchOnce (Display* dpy, Window w, int width, int atX, int atY,
    int kX, int kT, int kY, int grav, int delay)
{
    static int start_new = 1;
    static int t = 0;
    int d;

    if (start_new && !mono_p)
    {
	XColor fgc;
   
	fgc.red = random() % 65535;
	fgc.green = random() % 65535;
	fgc.blue = random() % 65535;
    
	if (XAllocColor(dpy, xgwa_colormap, &fgc))
	    XSetForeground(dpy, gc, fgc.pixel);
    }

    if (start_new)
    {
	t = 0;
    }

    for (d=0; d<delay && t<width; d++)
    {
	munchX(dpy, w, t, width, atX, atY, kX, kT, kY, grav);
	t++;
    }
    
    start_new = (t >= width) ? 1 : 0;
    
    return start_new;
}

/*
 * dumb way to get # of digits in number.  Probably faster than actually
 * doing a log and a division, maybe.
 */
static int dumb_log_2(int k) {
    int r = -1;
    while (k > 0) {
	k >>= 1; r++;
    }
    return r;
}

static int logmaxwidth;

static unsigned munch_instantiate(Display *dpy, Window w)
{
    int maxwidth;
    /*XWindowAttributes xgwa;*/
    Colormap cmap;
    XGCValues gcv;

    _dpy = dpy;

    /* get the dimensions of the window */
    /* XGetWindowAttributes (dpy, w, &xgwa); */
    {
	int screen_num;
	int dummy;
	void *pdummy = &dummy;
    
	XGetGeometry(dpy, w, pdummy, pdummy, pdummy, 
	    &xgwa_width, &xgwa_height, pdummy, pdummy);
	screen_num = DefaultScreen(dpy);
	xgwa_colormap = XDefaultColormap(dpy, screen_num);
    }
    
    /* We need a square; limit on screen size? */
    /* we want a power of 2 for the width or the munch doesn't fill up.
     */
    logmaxwidth = (int)
	dumb_log_2(xgwa_height < xgwa_width ? xgwa_height : xgwa_width);

    maxwidth = 1 << logmaxwidth;

    if (logmaxwidth < logminwidth) {
	/* off-by-one error here?  Anyone running on < 640x480? */
	fprintf(stderr, "munch: screen too small; use -logminwidth\n");
	fprintf(stderr, "\t(width is %d; log is %d; log must be at least "
		"%d)\n", 
		(xgwa_height < xgwa_width ? xgwa_height : xgwa_width),
		logmaxwidth, logminwidth);
	return FF_FAIL;
    }
    
    points = malloc(sizeof(XPoint) * (xgwa_height > xgwa_width ? xgwa_height : xgwa_width));
    if (points == NULL)
	return FF_FAIL;
    
    /* create the gc */
    cmap = xgwa_colormap;
    gcv.foreground= get_pixel_resource("foreground","Foreground",
				       dpy, cmap);
    gcv.background= get_pixel_resource("background","Background",
				       dpy, cmap);
    
    gc = XCreateGC(dpy, w, GCForeground|GCBackground, &gcv);
    
    gcv.foreground = gcv.background;
    cleargc = XCreateGC(dpy, w, GCForeground|GCBackground, &gcv);

    hold = get_integer_resource ("hold", "Integer");
    if (hold < 0) hold = 0;

    logminwidth = get_integer_resource ("logminwidth", "Integer");
    if (logminwidth < 2) logminwidth = 2;

    return FF_SUCCESS;
}

static unsigned munch_deinstantiate(void)
{
    free(points);
    XFreeGC(_dpy, gc);
    XFreeGC(_dpy, cleargc);
    return FF_SUCCESS;
}

/* This parses arguments, initializes the window, etc., and finally starts
 * calling munchOnce ad infinitum.
 */
void
screenhack_frame (Display *dpy, Window w)
{
    static int n = 0;  /* number of squares before we have to clear */
    static int randflags;
    static int thiswidth;
    static int atX, atY, kX, kT, kY, grav;
    static int start_new = 1;

    shiftk = get_boolean_resource("shift", "Boolean");    
    xor = get_boolean_resource("xor", "Boolean");

    /* always draw xor on mono. */
    if (mono_p || xor) {
	XSetFunction(dpy, gc, GXxor);
    }
    else
    {
	XSetFunction(dpy, gc, GXcopy);
    }

    clear = get_integer_resource ("clear", "Integer");
    if (clear < 0) clear = 0;

    delay = get_integer_resource ("delay", "Integer");
    if (delay < 0) delay = 0;

    if (start_new)
    {
	/* saves some calls to random.  big deal */
	randflags = random();
    
	/* choose size -- power of two, width, in pixels */
	thiswidth = 1 << (logminwidth + (random() % (1 + logmaxwidth - logminwidth)));
	
	/* draw at this location */
	atX = (random() % (xgwa_width <= thiswidth ? 1
			   : xgwa_width - thiswidth));
	atY = (random() % (xgwa_height <= thiswidth ? 1
			   : xgwa_width - thiswidth));

	/* wrap-around by these values; no need to %
		 as we end up doing that later anyway*/
	kX = ((shiftk && (randflags & SHIFT_KX))
	       ? (random() % thiswidth) : 0);
	kT = ((shiftk && (randflags & SHIFT_KT))
	       ? (random() % thiswidth) : 0);
	kY = ((shiftk && (randflags & SHIFT_KY))
	       ? (random() % thiswidth) : 0);
	
	/* set the gravity of the munch, or rather,
	 which direction we draw stuff in. */
	grav = (randflags & GRAV);
    }
    
    start_new = munchOnce(dpy, w, thiswidth, atX, atY, kX, kT, kY, grav, delay);

/*
    if (hold) usleep(hold);
*/  
    if (start_new)
	n++;
    if (clear && n >= clear)
    {
	/*XClearWindow(dpy, w);*/
	XFillRectangle(dpy, w, cleargc, 0, 0, xgwa_width, xgwa_height);
	n = 0;
    }
    XSync(dpy, False);
}
